/* Copyright (C) 2009, 2010, 2011 Keith Crane

This file is part DFILE Tools.

DFILE Tools is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at
your option) any later version.

DFILE Tools is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with DFILE Tools; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>. */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include "tbox.h"
#include "dfile.h"
#include "dfile_dynamic.h"
#include "dfile_agfunc.h"


#define NULL_CHECK( x ) ( ( *x == (const char *)0 ) ? "null" : *x )

static void	print_usage( const char * );

/*
** This function processes the command line arguments.
*/
int get_args( int argc, char * const argv[], const char **input_dfile_name, const char **output_dfile_name, int *thread_input_flag, int *thread_output_flag, dfile_tag_t **dfile_tag_tbl, unsigned short *dfile_tag_tbl_cnt, const char **sort_key, func_t **func_tbl, unsigned short *func_tbl_cnt, const char **input_filter_file_name, const char **output_filter_file_name )
{
	int	ch, ndx;
	extern char	*optarg;
	const char	*src_field, *result_field, *output_format;

	assert( argv != (char * const *) 0 );
	assert( input_dfile_name != (const char **) 0 );
	assert( output_dfile_name != (const char **) 0 );
	assert( thread_input_flag != (int *) 0 );
	assert( thread_output_flag != (int *) 0 );
	assert( dfile_tag_tbl != (dfile_tag_t **) 0 );
	assert( dfile_tag_tbl_cnt != (unsigned short *) 0 );
	assert( sort_key != (const char **)0 );
	assert( func_tbl != (func_t **)0 );
	assert( func_tbl_cnt != (unsigned short *)0 );
	assert( input_filter_file_name != (const char **)0 );
	assert( output_filter_file_name != (const char **)0 );

#ifndef NDEBUG
	/*
	** Initialize debug to off.
	*/
	Debug = 0;
#endif

	*input_dfile_name = (const char *)0;
	*output_dfile_name = (const char *)0;
	*dfile_tag_tbl = (dfile_tag_t *)0;
	*dfile_tag_tbl_cnt = (unsigned short)0;
	*sort_key = (const char *)0;
	*func_tbl = (func_t *)0;
	*func_tbl_cnt = (unsigned short)0;
	*input_filter_file_name = (const char *)0;
	*output_filter_file_name = (const char *)0;

	/*
	** Initialize I/O threading flags to false.
	*/
	*thread_input_flag = 0;
	*thread_output_flag = 0;

	while ( ( ch = getopt( argc, argv, "a:c:di:k:M:m:N:n:o:s:t:rwx:z:" ) ) != EOF ) {
		switch ( ch ) {
		case 'a':
			if ( parse_avg_sum_field_list( &src_field, &result_field, &output_format, optarg ) == -1 ) {
				return -1;
			}
			if ( alloc_func( func_tbl, func_tbl_cnt, Average, src_field, result_field, output_format ) == -1 ) {
				return -1;
			}
			break;
		case 'c':
			if ( alloc_func( func_tbl, func_tbl_cnt, Count, (const char *)0, optarg, (const char *)0 ) == -1 ) {
				return -1;
			}
			break;
		case 'd':
#ifdef NDEBUG
			(void) fputs( "Not compiled for debug.\n", stderr );
#else
			Debug = 1;
#endif
			break;
		case 'i':
			*input_dfile_name = optarg;
			break;
		case 'k':
			*sort_key = optarg;
			break;
		case 'M':
			if ( parse_field_list( &src_field, &result_field, optarg ) == -1 ) {
				return -1;
			}
			if ( alloc_func( func_tbl, func_tbl_cnt, Str_max, src_field, result_field, (const char *)0 ) == -1 ) {
				return -1;
			}
			break;
		case 'm':
			if ( parse_field_list( &src_field, &result_field, optarg ) == -1 ) {
				return -1;
			}
			if ( alloc_func( func_tbl, func_tbl_cnt, Str_min, src_field, result_field, (const char *)0 ) == -1 ) {
				return -1;
			}
			break;
		case 'N':
			if ( parse_field_list( &src_field, &result_field, optarg ) == -1 ) {
				return -1;
			}
			if ( alloc_func( func_tbl, func_tbl_cnt, Num_max, src_field, result_field, (const char *)0 ) == -1 ) {
				return -1;
			}
			break;
		case 'n':
			if ( parse_field_list( &src_field, &result_field, optarg ) == -1 ) {
				return -1;
			}
			if ( alloc_func( func_tbl, func_tbl_cnt, Num_min, src_field, result_field, (const char *)0 ) == -1 ) {
				return -1;
			}
			break;
		case 'o':
			*output_dfile_name = optarg;
			break;
		case 'r':
			/*
			** Set threaded input to true.
			*/
			*thread_input_flag = 1;
			break;
		case 's':
			if ( parse_avg_sum_field_list( &src_field, &result_field, &output_format, optarg ) == -1 ) {
				return -1;
			}
			if ( alloc_func( func_tbl, func_tbl_cnt, Sum, src_field, result_field, output_format ) == -1 ) {
				return -1;
			}
			break;
		case 't':
			if ( cmd_line_parse_tag( dfile_tag_tbl, dfile_tag_tbl_cnt, optarg ) == -1 ) {
				return -1;
			}
			break;
		case 'w':
			/*
			** Set threaded output to true.
			*/
			*thread_output_flag = 1;
			break;
		case 'x':
			*input_filter_file_name = optarg;
			break;
		case 'z':
			*output_filter_file_name = optarg;
			break;
		default:
			print_usage( argv[ 0 ] );
			return -1;
		}
	}

	if ( *input_dfile_name == (char *) 0 ) {
		FPUT_SRC_CODE( stderr );
		(void) fputs( "Must specify input dfile name.\n", stderr );
		print_usage( argv[ 0 ] );
		return -1;
	}

	if ( *output_dfile_name == (char *) 0 ) {
		FPUT_SRC_CODE( stderr );
		(void) fputs( "Must specify output dfile name.\n", stderr );
		print_usage( argv[ 0 ] );
		return -1;
	}

	if ( *func_tbl_cnt == (unsigned short) 0 ) {
		FPUT_SRC_CODE( stderr );
		(void) fputs( "Must specify at least one aggregate function.\n", stderr );
		print_usage( argv[ 0 ] );
		return -1;
	}

	if ( Debug ) {
		(void) fprintf( stderr, "Input DFile Name: [%s]\n", NULL_CHECK( input_dfile_name ) );
		(void) fprintf( stderr, "Output DFile Name: [%s]\n", NULL_CHECK( output_dfile_name ) );
		(void) fprintf( stderr, "Input Filter File Name: [%s]\n", NULL_CHECK( input_filter_file_name ) );
		(void) fprintf( stderr, "Output Filter File Name: [%s]\n", NULL_CHECK( output_filter_file_name ) );
		(void) fprintf( stderr, "Thread Input Flag: [%s]\n", ( *thread_input_flag ) ? "true" : "false" );
		(void) fprintf( stderr, "Thread Output Flag: [%s]\n", ( *thread_output_flag ) ? "true" : "false" );
		(void) fputs( "Input files:\n", stderr );
		for ( ndx = (unsigned short)0; ndx < *dfile_tag_tbl_cnt; ++ndx ) {
			(void) fprintf( stderr, "DFile Tag [%s] [%s]\n", ( *dfile_tag_tbl )[ ndx ].tag, ( *dfile_tag_tbl )[ ndx ].tag_value );
		}
		(void) fprintf( stderr, "Sort Key: [%s]\n", NULL_CHECK( sort_key ) );
	}

	return 0;
}

static void print_usage( const char *exec_name )
{
	(void) fputs( "usage: ", stderr );
	(void) fputs( exec_name, stderr );
	(void) fputs( " [-d]", stderr );
	(void) fputs( " -a field,field[,%g]", stderr );
	(void) fputs( " -c field", stderr );
	(void) fputs( " -i dfile", stderr );
	(void) fputs( " [-k field,field...]", stderr );
	(void) fputs( " -M field,field", stderr );
	(void) fputs( " -m field,field", stderr );
	(void) fputs( " -N field,field", stderr );
	(void) fputs( " -n field,field", stderr );
	(void) fputs( " -o dfile", stderr );
	(void) fputs( " -s field,field[,%g]", stderr );
	(void) fputs( " [-t %x=abc]", stderr );
	(void) fputs( " [-r]", stderr );
	(void) fputs( " [-w]", stderr );
	(void) fputs( " [-x file]", stderr );
	(void) fputs( " [-z file]", stderr );
	(void) fputc( '\n', stderr );
	(void) fputs( "\t-d -> debug\n", stderr );
	(void) fputs( "\t-a -> average source,result field\n", stderr );
	(void) fputs( "\t-c -> record count result field\n", stderr );
	(void) fputs( "\t-i -> input DFile\n", stderr );
	(void) fputs( "\t-k -> key sort fields\n", stderr );
	(void) fputs( "\t-M -> ASCII maximun source,result field\n", stderr );
	(void) fputs( "\t-m -> ASCII minimum source,result field\n", stderr );
	(void) fputs( "\t-N -> numeric maximum source,result field\n", stderr );
	(void) fputs( "\t-n -> numeric minimum source,result field\n", stderr );
	(void) fputs( "\t-o -> output DFile\n", stderr );
	(void) fputs( "\t-t -> DFile path tags\n", stderr );
	(void) fputs( "\t-r -> thread input (default no threading)\n", stderr );
	(void) fputs( "\t-s -> sum source,result field\n", stderr );
	(void) fputs( "\t-w -> thread output (default no threading)\n", stderr );
	(void) fputs( "\t-x -> input filter file\n", stderr );
	(void) fputs( "\t-z -> output filter file\n", stderr );
}
